/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

var ActionStatus;
(function (ActionStatus) {
    ActionStatus["BUSY"] = "busy";
    ActionStatus["SUCCESS"] = "success";
    ActionStatus["FAILURE"] = "failure";
})(ActionStatus || (ActionStatus = {}));
const values = Object.values(ActionStatus);
/* istanbul ignore next */
(function (ActionStatus) {
    function has(value) {
        return values.includes(value);
    }
    ActionStatus.has = has;
})(ActionStatus || (ActionStatus = {}));

const prepareChain = (chain) => typeof chain === 'function'
    ? (payload) => __awaiter(void 0, void 0, void 0, function* () { return [payload, yield chain(payload)]; })
    : (payload) => __awaiter(void 0, void 0, void 0, function* () { return [payload, yield chain]; });
const createDispatcher = (store, next, extra, type) => ([payload, chain]) => {
    if (chain) {
        if (chain instanceof Array) {
            for (let i = 0; i < chain.length; i++) {
                if (chain[i])
                    store.dispatch(chain[i]);
            }
        }
        else {
            store.dispatch(chain);
        }
    }
    return next(Object.assign(Object.assign({}, extra), { payload, type }));
};
const createErrorHandler = (store, next, failureType, dispatchOnError, action, extra) => (error) => {
    const payload = {
        error,
        origin: action,
    };
    const nextval = next(Object.assign(Object.assign({}, extra), { type: failureType, error: true, payload }));
    if (dispatchOnError) {
        if (typeof dispatchOnError === 'string') {
            store.dispatch({
                type: dispatchOnError,
                error: true,
                payload,
            });
        }
        else {
            const errorAction = dispatchOnError(payload);
            if (errorAction)
                store.dispatch(errorAction);
        }
    }
    return nextval;
};
const createPromiseResolver = (store) => (next, promise, action, successType, failureType, chain, dispatchOnError, extra) => promise
    .then(prepareChain(chain))
    .then(createDispatcher(store, next, extra, successType))
    .catch(createErrorHandler(store, next, failureType, dispatchOnError, action, extra));

const HOP = Object.prototype.hasOwnProperty;
const createMiddleware = (store, next, resolve) => (action) => {
    if (!HOP.call(action, 'promise')) {
        if (typeof action.transform !== 'function')
            return next(action);
        return next(action.transform(action, store.getState()));
    }
    const { type, error, payload, promise, statusSelector, chain, dispatchOnError, transform } = action, extra = __rest(action, ["type", "error", "payload", "promise", "statusSelector", "chain", "dispatchOnError", "transform"]);
    if (action.error === true)
        return next(action);
    if (statusSelector) { // Blocking  action
        if (!(type instanceof Array)) {
            throw new Error('`type` must be Array if `statusSelector` is defined');
        }
        const state = store.getState();
        if (!(statusSelector instanceof Array)) { // Single status selector
            const currentStatus = statusSelector(state);
            if (currentStatus === ActionStatus.BUSY || currentStatus === false) {
                return null;
            }
        }
        else { // Block action if any status selector returns BUSY
            for (let i = 0; i < statusSelector.length; i++) {
                const currentStatus = statusSelector[i](state);
                if (currentStatus === ActionStatus.BUSY || currentStatus === false) {
                    return null;
                }
            }
        }
    }
    // If action.transform function is set, apply it before passing action to next.
    const _next = typeof transform === 'function'
        ? ((action) => next(transform(action, store.getState())))
        : next;
    let successType;
    let failureType;
    if (type instanceof Array) {
        _next(Object.assign({ type: type[0], payload }, extra)); // Dispatch start action
        successType = type[1];
        failureType = type[2];
    }
    else {
        successType = failureType = type;
    }
    // At this point, start action has been reduced and status has been updated if necessary
    const _promise = typeof promise === 'function' ? promise() : promise;
    return resolve(_next, _promise, action, successType, failureType, chain, dispatchOnError, extra);
};
const middleware = (store) => (next) => createMiddleware(store, next, createPromiseResolver(store));

const createThunklessAction = (action) => action;

export { ActionStatus, createThunklessAction, middleware };
